"""
Tests for panel widgets and param objects as arguments
"""
from unittest import TestCase, SkipTest

from hvplot.util import process_xarray  # noqa

def look_for_class(panel, classname, items=None):
    """
    Descend a panel object and find any instances of the given class
    """
    import panel as pn

    if items is None:
        items = []
    if isinstance(panel, pn.layout.ListPanel):
        for p in panel:
            items = look_for_class(p, classname, items)
    elif isinstance(panel, classname):
        items.append(panel)
    return items


class TestPanelObjects(TestCase):

    def setUp(self):
        try:
            import panel as pn  # noqa
            import hvplot.pandas  # noqa
        except:
            raise SkipTest('panel not available')

        from bokeh.sampledata.iris import flowers
        self.flowers = flowers
        self.cols = list(self.flowers.columns[:-1])


    def test_using_explicit_widgets_works(self):
        import panel as pn

        x = pn.widgets.Select(name='x', value='sepal_length', options=self.cols)
        y = pn.widgets.Select(name='y', value='sepal_width', options=self.cols)
        kind = pn.widgets.Select(name='kind', value='scatter', options=['bivariate', 'scatter'])
        by_species = pn.widgets.Checkbox(name='By species')
        color = pn.widgets.ColorPicker(value='#ff0000')

        @pn.depends(by_species.param.value, color.param.value)
        def by_species_fn(by_species, color):
            return 'species' if by_species else color

        self.flowers.hvplot(x, y=y, kind=kind.param.value, c=color)

    def test_casting_widgets_to_different_classes(self):
        import panel as pn

        pane = self.flowers.hvplot.scatter(
            groupby='species', legend='top_right',
            widgets={'species': pn.widgets.DiscreteSlider})

        assert len(look_for_class(pane, pn.widgets.DiscreteSlider)) == 1

    def test_using_explicit_widgets_with_groupby_does_not_raise_error(self):
        import panel as pn

        x = pn.widgets.Select(name='x', value='sepal_length', options=self.cols)
        y = pn.widgets.Select(name='y', value='sepal_width', options=self.cols)

        pane = self.flowers.hvplot(x, y, groupby='species')
        assert isinstance(pane, pn.param.ParamFunction)



def patch(name='hvplot', extension='bokeh', logo=False):
    from . import hvPlotTabular, post_patch

    try:
        import streamz.dataframe as sdf
    except ImportError:
        raise ImportError('Could not patch plotting API onto streamz. '
                          'Streamz could not be imported.')
    _patch_plot = lambda self: hvPlotTabular(self)
    _patch_plot.__doc__ = hvPlotTabular.__call__.__doc__
    patch_property = property(_patch_plot)
    setattr(sdf.DataFrame, name, patch_property)
    setattr(sdf.DataFrames, name, patch_property)
    setattr(sdf.Series, name, patch_property)
    setattr(sdf.Seriess, name, patch_property)

    post_patch(extension, logo)

patch()



from unittest import TestCase

import pandas as pd

from holoviews.streams import Buffer, Pipe


class TestExplicitStreamPlotting(TestCase):
    
    def setUp(self):
        import hvplot.pandas   # noqa
        self.df = pd.DataFrame([[1, 2], [3, 4], [5, 6]], columns=['x', 'y'])

    def test_pipe_stream(self):
        stream = Pipe(data=self.df)
        plot = self.df.hvplot('x', 'y', stream=stream)
        pd.testing.assert_frame_equal(plot[()].data, self.df)
        new_df = pd.DataFrame([[7, 8], [9, 10]], columns=['x', 'y'])
        stream.send(new_df)
        pd.testing.assert_frame_equal(plot[()].data, new_df)

    def test_buffer_stream(self):
        stream = Buffer(data=self.df, index=False)
        plot = self.df.hvplot('x', 'y', stream=stream)
        pd.testing.assert_frame_equal(plot[()].data, self.df)
        new_df = pd.DataFrame([[7, 8], [9, 10]], columns=['x', 'y'])
        stream.send(new_df)
        pd.testing.assert_frame_equal(plot[()].data, pd.concat([self.df, new_df]))



"""
Tests pandas.options.backend setting
"""
from packaging.version import Version
from unittest import TestCase, SkipTest
import pandas as pd

from parameterized import parameterized

from ..converter import HoloViewsConverter


no_args = ['line', 'area', 'hist', 'box', 'kde', 'density', 'bar', 'barh']
x_y = ['scatter', 'hexbin']

no_args_mapping = [(kind, el) for kind, el in HoloViewsConverter._kind_mapping.items()
                   if kind in no_args]
x_y_mapping = [(kind, el) for kind, el in HoloViewsConverter._kind_mapping.items()
               if kind in x_y]


class TestPandasHoloviewsPlotting(TestCase):

    def setUp(self):
        if Version(pd.__version__) < Version('0.25.1'):
            raise SkipTest('entrypoints for plotting.backends was added '
                           'in pandas 0.25.1')
        pd.options.plotting.backend = 'holoviews'

    @parameterized.expand(no_args_mapping)
    def test_pandas_series_plot_returns_holoviews_object(self, kind, el):
        series = pd.Series([0, 1, 2])
        plot = getattr(series.plot, kind)()
        self.assertIsInstance(plot, el)

    @parameterized.expand(no_args_mapping)
    def test_pandas_dataframe_plot_returns_holoviews_object(self, kind, el):
        df = pd.DataFrame([0, 1, 2])
        plot = getattr(df.plot, kind)()
        self.assertIsInstance(plot, el)

    @parameterized.expand(x_y_mapping)
    def test_pandas_dataframe_plot_returns_holoviews_object_when_x_and_y_set(self, kind, el):
        df = pd.DataFrame({'a': [0, 1, 2], 'b': [5, 7, 2]})
        plot = getattr(df.plot, kind)(x='a', y='b')
        self.assertIsInstance(plot, el)

    def test_pandas_dataframe_plot_does_not_implement_pie(self):
        df = pd.DataFrame({'a': [0, 1, 2], 'b': [5, 7, 2]})
        with self.assertRaisesRegex(NotImplementedError, "pie"):
            df.plot.pie(y='a')


class TestPandasHvplotPlotting(TestPandasHoloviewsPlotting):

    def setUp(self):
        if Version(pd.__version__) < Version('0.25.1'):
            raise SkipTest('entrypoints for plotting.backends was added '
                           'in pandas 0.25.1')
        pd.options.plotting.backend = 'hvplot'



import inspect
import textwrap

from functools import wraps as _wraps

import param
import panel as _pn
import holoviews as _hv

from holoviews import Store, render  # noqa

from .converter import HoloViewsConverter
from .interactive import Interactive
from .utilities import hvplot_extension, output, save, show # noqa
from .plotting import (hvPlot, hvPlotTabular,  # noqa
                       andrews_curves, lag_plot,
                       parallel_coordinates, scatter_matrix, plot)

__version__ = str(param.version.Version(fpath=__file__, archive_commit="$Format:%h$",
                                        reponame="hvplot"))

_METHOD_DOCS = {}

def _get_doc_and_signature(
    cls, kind, completions=False, docstring=True, generic=True, style=True, signature=None
):
    converter = HoloViewsConverter
    method = getattr(cls, kind)
    kind_opts = converter._kind_options.get(kind, [])
    eltype = converter._kind_mapping[kind]

    formatter = ''
    if completions:
        formatter = "hvplot.{kind}({completions})"
    if docstring:
        if formatter:
            formatter += '\n'
        formatter += "{docstring}"
    if generic:
        if formatter:
            formatter += '\n'
        formatter += "{options}"

    # Bokeh is the default backend
    backend = hvplot_extension.compatibility or Store.current_backend
    if eltype in Store.registry[backend]:
        valid_opts = Store.registry[backend][eltype].style_opts
        if style:
            formatter += '\n{style}'
    else:
        valid_opts = []

    style_opts = 'Style options\n-------------\n\n' + '\n'.join(sorted(valid_opts))

    parameters = []
    extra_kwargs = _hv.core.util.unique_iterator(
        valid_opts + kind_opts + converter._axis_options + converter._op_options
    )

    sig = signature or inspect.signature(method)
    for name, p in list(sig.parameters.items())[1:]:
        if p.kind == 1:
            parameters.append((name, p.default))

    filtered_signature = [p for p in sig.parameters.values()
                            if p.kind != inspect.Parameter.VAR_KEYWORD]
    extra_params = [inspect.Parameter(k, inspect.Parameter.KEYWORD_ONLY)
                    for k in extra_kwargs
                    if k not in [p.name for p in filtered_signature]]
    all_params = (filtered_signature + extra_params
                    + [inspect.Parameter('kwargs', inspect.Parameter.VAR_KEYWORD)])
    signature = inspect.Signature(all_params)

    parameters += [(o, None) for o in extra_kwargs]
    completions = ', '.join([f'{n}={v}' for n, v in parameters])
    options = textwrap.dedent(converter.__doc__)
    method_doc = _METHOD_DOCS.get(kind, method.__doc__)
    _METHOD_DOCS[kind] = method_doc
    docstring = formatter.format(
        kind=kind, completions=completions, docstring=textwrap.dedent(method_doc),
        options=options, style=style_opts)
    return docstring, signature


def help(kind=None, docstring=True, generic=True, style=True):
    """
    Provide a docstring with all valid options which apply to the plot
    type.

    Parameters
    ----------
    kind: str
        The kind of plot to provide help for
    docstring: boolean (default=True)
        Whether to display the docstring
    generic: boolean (default=True)
        Whether to provide list of generic options
    style: boolean (default=True)
        Whether to provide list of style options
    """
    doc, sig = _get_doc_and_signature(cls=hvPlot, kind=kind,
                                      docstring=docstring, generic=generic, style=style)
    print(doc)


def post_patch(extension='bokeh', logo=False):
    if extension and not getattr(_hv.extension, '_loaded', False):
        if getattr(_pn.extension, '_loaded', False):
            ext = _hv.extension._backends[extension]
            __import__('holoviews.plotting.%s' % ext)
            _hv.Store.set_current_backend(extension)
        else:
            hvplot_extension(extension, logo=logo)


def _patch_doc(cls, kind, signature=None):
    method = getattr(cls, kind)
    docstring, signature = _get_doc_and_signature(cls, kind, False, signature=signature)
    method.__doc__ = docstring
    method.__signature__ = signature


class _PatchHvplotDocstrings:

    def __init__(self):
        # Store the original signatures because the method signatures
        # are going to be patched everytime an extension is changed.
        signatures = {}
        for cls in [hvPlot, hvPlotTabular]:
            for _kind in HoloViewsConverter._kind_mapping:
                if hasattr(cls, _kind):
                    method = getattr(cls, _kind)
                    sig = inspect.signature(method)
                    signatures[(cls, _kind)] = sig
        self.orig_signatures = signatures

    def __call__(self):
        for cls in [hvPlot, hvPlotTabular]:
            for _kind in HoloViewsConverter._kind_mapping:
                if hasattr(cls, _kind):
                    signature = self.orig_signatures[(cls, _kind)]
                    _patch_doc(cls, _kind, signature=signature)


_patch_hvplot_docstrings = _PatchHvplotDocstrings()
_patch_hvplot_docstrings()

def _hook_patch_docstrings(backend):
    # Patch or re-patch the docstrings/signatures to display
    # the right styling options.
    from . import _patch_hvplot_docstrings
    _patch_hvplot_docstrings()

Store._backend_switch_hooks.append(_hook_patch_docstrings)

extension = hvplot_extension

@_wraps(_pn.bind)
def bind(function, *args, **kwargs):
    bound = _pn.bind(function, *args, **kwargs)
    bound.interactive = lambda **kwargs: Interactive(bound, **kwargs)
    return bound



"""
Set of transforms to go from a Bokeh option to another backend's option.
"""
from holoviews.core.overlay import CompositeOverlay
from holoviews.core.options import Store
from holoviews.plotting.util import COLOR_ALIASES


UNSET = type('UNSET', (), {})


def _transform_size_to_mpl(width, height, aspect):
    opts = {}
    if width and height:
        opts = {'aspect': width/height, 'fig_size': (width/300.)*100}
    elif aspect and width:
        opts = {'aspect': aspect, 'fig_size': (width/300.)*100}
    elif aspect and height:
        opts = {'aspect': aspect, 'fig_size': (height/300.)*100}
    elif width:
        opts = {'fig_size': (width/300.)*100}
    elif height:
        opts = {'fig_size': (height/300.)*100}
    return opts


def _transfer_opts(element, backend):
    """
    Transfer the bokeh options of an element to another backend
    based on an internal mapping of option transforms.
    """
    elname = type(element).__name__
    options = Store.options(backend=backend)
    transforms = BACKEND_TRANSFORMS[backend]
    if isinstance(element, CompositeOverlay):
        element = element.apply(
            _transfer_opts, backend=backend, per_element=True
        )
    new_opts = {}
    el_options = element.opts.get(backend='bokeh', defaults=False).kwargs
    for grp, el_opts in options[elname].groups.items():
        for opt, val in el_options.items():
            transform = transforms.get(grp, {}).get(opt, None)
            # This condition could be applied to matplotlib only
            # but is applied to plotly too since there seems to be
            # no interactive options available like for the bokeh backend
            if transform is None and _is_interactive_opt(opt):
                transform = UNSET
            if transform is UNSET:
                continue
            elif transform:
                opt, val = transform(opt, val)
                if val is UNSET:
                    continue
            if opt not in el_opts.allowed_keywords:
                continue
            new_opts[opt] = val
    if backend == 'matplotlib':
        size_opts = _transform_size_to_mpl(
            el_options.get('width'), el_options.get('height'),
            el_options.get('aspect')
        )
        new_opts.update(size_opts)
    return element.opts(**new_opts, backend=backend)


def _is_interactive_opt(bk_opt):
    """
    Heuristics to detect if a bokeh option is about interactivity, like
    'selection_alpha'.

    >>> is_interactive_opt('height')
    False
    >>> is_interactive_opt('annular_muted_alpha')
    True
    """
    interactive_flags = [
        'hover',
        'muted',
        'nonselection',
        'selection',
    ]
    return any(part in interactive_flags for part in bk_opt.split('_'))


def _transfer_opts_cur_backend(element):
    if Store.current_backend != 'bokeh':
        element = element.apply(_transfer_opts, backend=Store.current_backend)
    return element


# Matplotlib transforms

_line_cap_bk_mpl_mapping = {
    'butt': 'butt',
    'round': 'round',
    'square': 'projecting',
}

_line_dash_bk_mpl_mapping = {
    'solid': 'solid',
    'dashed': 'dashed',
    'dotted': 'dotted',
    'dotdash': UNSET,
    'dashdot': 'dashdot',
}

_text_baseline_bk_mpl_mapping = {
    'top': 'top',
    'middle': 'center',
    'bottom': 'bottom',
    'alphabetic': UNSET,
    'hanging': UNSET,
    'ideographic': UNSET,
}

MATPLOTLIB_TRANSFORMS = {
    'plot': {
        'height': UNSET,
        'width': UNSET,
        'min_height': UNSET,
        'min_width': UNSET,
        'max_height': UNSET,
        'max_width': UNSET,
        'frame_width': UNSET,
        'frame_height': UNSET,
        'sizing_mode': UNSET,
        'responsive': UNSET,
        'shared_axes': UNSET,
        'shared_datasource': UNSET,
        'selected': UNSET,
        'tools': UNSET,
        'active_tools': UNSET,
        'default_tools': UNSET,
        'toolbar': UNSET,
        'align': UNSET,
        'lod': UNSET,
        'margin': UNSET,
        'border': UNSET,
        'jitter': UNSET,
        'legend_muted': UNSET,
        'legend_offset': UNSET,
        'gridstyle': UNSET,
        'colorbar_position': UNSET,
        'violin_width': UNSET,
    },
    'style': {
        'bar_width': UNSET,
        'size': lambda k, v: ('s', v),
        'fill_color': lambda k, v: ('facecolor', v),
        'fill_alpha': UNSET,
        'line_alpha': UNSET,
        'line_cap': lambda k, v: ('capstyle', _line_cap_bk_mpl_mapping.get(v, UNSET)),
        'line_color': lambda k, v: ('edgecolor', v),
        'line_dash': lambda k, v: ('linestyle', _line_dash_bk_mpl_mapping.get(v, UNSET)),
        'line_join': lambda k, v: ('joinstyle', v),
        'line_width': lambda k, v: ('linewidth', v),
        'palette': UNSET,
        'click_policy': UNSET,
        'annular_alpha': UNSET,
        'annular_color': UNSET,
        'annular_fill_alpha': UNSET,
        'annular_fill_color': UNSET,
        'annular_line_alpha': UNSET,
        'annular_line_cap': UNSET,
        'annular_line_color': lambda k, v: ('annular_edgecolors', v),
        'annular_line_dash': UNSET,
        'annular_line_join': UNSET,
        'annular_line_width': lambda k, v: ('annular_linewidth', v),
        'annular_visible': UNSET,
        'dilate': UNSET,
        'ticks_text_align': UNSET,
        'ticks_text_alpha': UNSET,
        'ticks_text_baseline': UNSET,
        'ticks_text_color': UNSET,
        'ticks_text_font': UNSET,
        'ticks_text_font_size': UNSET,
        'ticks_text_font_style': UNSET,
        'xmarks_alpha': UNSET,
        'xmarks_color': UNSET,
        'xmarks_line_alpha': UNSET,
        'xmarks_line_cap': UNSET,
        'xmarks_line_color': lambda k, v: ('xmarks_edgecolor', v),
        'xmarks_line_dash': UNSET,
        'xmarks_line_join': UNSET,
        'xmarks_line_width': lambda k, v: ('xmarks_linewidth', v),
        'xmarks_visible': UNSET,
        'ymarks_alpha': UNSET,
        'ymarks_color': UNSET,
        'ymarks_line_alpha': UNSET,
        'ymarks_line_cap': UNSET,
        'ymarks_line_color': lambda k, v: ('ymarks_edgecolor', v),
        'ymarks_line_dash': UNSET,
        'ymarks_line_join': UNSET,
        'ymarks_line_width': lambda k, v: ('ymarks_linewidth', v),
        'ymarks_visible': UNSET,
        'lower_head': lambda k, v: ('lolims', v),
        'upper_head': lambda k, v: ('uplims', v),
        'whisker_color': UNSET,
        'outlier_color': UNSET,
        'text_align': lambda k, v: ('horizontalalignment', v),
        'text_alpha': lambda k, v: ('alpha', v),
        'text_baseline': lambda k, v: ('verticalalignment', _text_baseline_bk_mpl_mapping.get(v, UNSET)),
        'text_color': lambda k, v: ('color', v),
        'text_font': UNSET, 
        'text_font_size': lambda k, v: ('size', v),
        'text_font_style': UNSET,
        'box_alpha': UNSET,
        'box_cmap': UNSET,
        'box_fill_alpha': UNSET,
        'box_fill_color': UNSET,
        'box_line_alpha': UNSET,
        'box_line_cap': UNSET,
        'box_line_color': UNSET,
        'box_line_dash': UNSET,
        'box_line_join': UNSET,
        'box_line_width': UNSET,
        'box_visible': UNSET,
        'median_alpha': UNSET,
        'median_color': UNSET,
        'outline_alpha': UNSET,
        'outline_color': UNSET,
        'outline_line_alpha': UNSET,
        'outline_line_cap': UNSET,
        'outline_line_color': UNSET,
        'outline_line_dash': UNSET,
        'outline_line_join': UNSET,
        'outline_line_width': UNSET,
        'outline_visible': UNSET,
        'stats_alpha': UNSET,
        'stats_line_alpha': UNSET,
        'stats_line_cap': UNSET,
        'stats_line_color': UNSET,
        'stats_line_dash': UNSET,
        'stats_line_join': UNSET,
        'stats_line_width': UNSET,
        'stats_visible': UNSET,
        'violin_alpha': UNSET,
        'violin_cmap': UNSET,
        'violin_color': UNSET,
        'violin_fill_alpha': UNSET,
        'violin_fill_color': UNSET,
        'violin_line_alpha': UNSET,
        'violin_line_cap': UNSET,
        'violin_line_color': UNSET,
        'violin_line_dash': UNSET,
        'violin_line_join': UNSET,
        'violin_line_width': UNSET,
        'violin_visible': UNSET,
        'editable': UNSET,
        'fit_columns': UNSET,
        'index_position': UNSET,
        'row_headers': UNSET,
        'scroll_to_selection': UNSET,
        'selectable': UNSET,
        'sortable': UNSET,
        'edge_alpha': lambda k, v: ('edge_alpha', v),
        'edge_cmap': lambda k, v: ('edge_cmap', v),
        'edge_color': lambda k, v: ('edge_color', v),
        'edge_fill_alpha': UNSET,
        'edge_fill_color': UNSET,
        'edge_line_alpha': UNSET,
        'edge_line_cap': UNSET,
        'edge_line_color': lambda k, v: ('edge_color', v),
        'edge_line_dash': lambda k, v: ('edge_linestyle', v),
        'edge_line_join': UNSET,
        'edge_line_width': lambda k, v: ('edge_linewidth', v),
        'edge_visible': UNSET,  
        'node_alpha': lambda k, v: ('node_alpha', v), 
        'node_cmap': lambda k, v: ('node_cmap', v), 
        'node_color': lambda k, v: ('node_color', v), 
        'node_fill_alpha': UNSET, 
        'node_fill_color': lambda k, v: ('node_facecolors', v), 
        'node_line_alpha': UNSET, 
        'node_line_cap': UNSET, 
        'node_line_color': lambda k, v: ('node_edgecolors', v), 
        'node_line_dash': UNSET, 
        'node_line_join': UNSET, 
        'node_line_width': lambda k, v: ('node_linewidth', v), 
        'node_marker': lambda k, v: ('node_marker', v), 
        'node_radius': UNSET, 
        'node_size': lambda k, v: ('node_size', v), 
        'node_visible': lambda k, v: ('visible', v),
    },
}

# Plotly transforms

_line_dash_bk_plotly_mapping = {
    'solid': 'solid',
    'dashed': 'dash',
    'dotted': 'dot',
    'dotdash': UNSET,
    'dashdot': 'dashdot',
}

PLOTLY_TRANSFORMS = {
    'plot': {
        'alpha': lambda k, v: ('opacity', v),
        'min_height': UNSET,
        'min_width': UNSET,
        'max_height': UNSET,
        'max_width': UNSET,
        'frame_width': UNSET,
        'frame_height': UNSET,
        'batched': UNSET,
        'legend_limit': UNSET,
        'tools': UNSET,
        'shared_xaxis': UNSET,
        'shared_yaxis': UNSET,
    },
    'style': {
        'alpha': lambda k, v: ('opacity', v),
        'color': lambda k, v: (k, COLOR_ALIASES.get(v, v)),
        'fill_color': lambda k, v: ('fillcolor', COLOR_ALIASES.get(v, v)),
        'line_color': lambda k, v: (k, COLOR_ALIASES.get(v, v)),
        'line_dash': lambda k, v: ('dash', _line_dash_bk_plotly_mapping.get(v, UNSET)),
        'line_width': lambda k, v: ('line_width', v),
        'muted_alpha': UNSET,
        'palette': UNSET,
    }
}


BACKEND_TRANSFORMS = dict(
    matplotlib=MATPLOTLIB_TRANSFORMS,
    plotly=PLOTLY_TRANSFORMS,
)



from collections import OrderedDict
from unittest import SkipTest

import numpy as np
import pandas as pd

from hvplot.plotting import hvPlot, hvPlotTabular
from holoviews import Store, Scatter
from holoviews.element.comparison import ComparisonTestCase

try:
    import xarray as xr
except:
    xr = None


class TestOverrides(ComparisonTestCase):

    def setUp(self):
        import hvplot.pandas  # noqa
        self.df = pd.DataFrame([[1, 2], [3, 4], [5, 6]], columns=['x', 'y'])

    def test_define_default_options(self):
        hvplot = hvPlotTabular(self.df, width=42, height=42)
        curve = hvplot(y='y')
        opts = Store.lookup_options('bokeh', curve, 'plot')
        self.assertEqual(opts.options.get('width'), 42)
        self.assertEqual(opts.options.get('height'), 42)

    def test_define_custom_method(self):
        hvplot = hvPlotTabular(self.df, {'custom_scatter': {'width': 42, 'height': 42}})
        custom_scatter = hvplot.custom_scatter(y='y')
        scatter = hvplot.scatter(y='y')
        custom_opts = Store.lookup_options('bokeh', custom_scatter, 'plot')
        opts = Store.lookup_options('bokeh', scatter, 'plot')
        self.assertEqual(custom_opts.options.get('width'), 42)
        self.assertEqual(custom_opts.options.get('height'), 42)
        self.assertNotEqual(opts.options.get('width'), 42)
        self.assertNotEqual(opts.options.get('height'), 42)

    def test_define_customize_method(self):
        hvplot = hvPlotTabular(self.df, {'scatter': {'width': 42, 'height': 42}})
        custom_scatter = hvplot.scatter(y='y')
        curve = hvplot.line(y='y')
        custom_opts = Store.lookup_options('bokeh', custom_scatter, 'plot')
        opts = Store.lookup_options('bokeh', curve, 'plot')
        self.assertEqual(custom_opts.options.get('width'), 42)
        self.assertEqual(custom_opts.options.get('height'), 42)
        self.assertNotEqual(opts.options.get('width'), 42)
        self.assertNotEqual(opts.options.get('height'), 42)

    def test_attempt_to_override_kind_on_method(self):
        hvplot = hvPlotTabular(self.df, {'scatter': {'kind': 'line'}})
        self.assertIsInstance(hvplot.scatter(y='y'), Scatter)

    def test_pandas_query_metadata(self):
        hvplot = hvPlotTabular(self.df, query='x>2')
        assert len(hvplot._data) == 2


class TestXArrayOverrides(ComparisonTestCase):

    def setUp(self):
        if xr is None:
            raise SkipTest('XArray not available')
        coords = OrderedDict([('time', [0, 1]), ('lat', [0, 1]), ('lon', [0, 1])])
        self.da_img_by_time = xr.DataArray(np.arange(8).reshape((2, 2, 2)),
                                           coords, ['time', 'lat', 'lon']).assign_coords(
                                               lat1=xr.DataArray([2,3], dims=['lat']))

    def test_xarray_isel_scalar_metadata(self):
        hvplot = hvPlot(self.da_img_by_time, isel={'time': 1})
        assert hvplot._data.ndim == 2

    def test_xarray_isel_nonscalar_metadata(self):
        hvplot = hvPlot(self.da_img_by_time, isel={'time': [1]})
        assert hvplot._data.ndim == 3
        assert len(hvplot._data.time) == 1

    def test_xarray_sel_metadata(self):
        hvplot = hvPlot(self.da_img_by_time, sel={'time': 1})
        assert hvplot._data.ndim == 2



import sys

from unittest import SkipTest
from parameterized import parameterized

import numpy as np
import pandas as pd

from holoviews import Store
from holoviews.element import Image, QuadMesh
from holoviews.element.comparison import ComparisonTestCase


class TestDatashader(ComparisonTestCase):

    def setUp(self):
        try:
            import datashader # noqa
        except:
            raise SkipTest('Datashader not available')
        if sys.maxsize < 2**32:
            raise SkipTest('Datashader does not support 32-bit systems')
        import hvplot.pandas # noqa
        self.df = pd.DataFrame([[1, 2, 'A', 0.1], [3, 4, 'B', 0.2], [5, 6, 'C', 0.3]],
                               columns=['x', 'y', 'category', 'number'])

    def test_rasterize_by_cat(self):
        from datashader.reductions import count_cat
        dmap = self.df.hvplot.scatter('x', 'y', by='category', rasterize=True)
        agg = dmap.callback.inputs[0].callback.operation.p.aggregator
        self.assertIsInstance(agg, count_cat)
        self.assertEqual(agg.column, 'category')

    @parameterized.expand([('rasterize',), ('datashade',)])
    def test_color_dim_with_default_agg(self, operation):
        from datashader.reductions import mean
        dmap = self.df.hvplot.scatter('x', 'y', c='number', **{operation: True})
        agg = dmap.callback.inputs[0].callback.operation.p.aggregator
        self.assertIsInstance(agg, mean)
        self.assertEqual(agg.column, 'number')

    @parameterized.expand([('rasterize',), ('datashade',)])
    def test_color_dim_with_string_agg(self, operation):
        from datashader.reductions import sum
        dmap = self.df.hvplot.scatter('x', 'y', c='number', aggregator='sum', **{operation: True})
        agg = dmap.callback.inputs[0].callback.operation.p.aggregator
        self.assertIsInstance(agg, sum)
        self.assertEqual(agg.column, 'number')

    @parameterized.expand([('rasterize',), ('datashade',)])
    def test_color_dim_also_an_axis(self, operation):
        from datashader.reductions import mean
        original_data = self.df.copy(deep=True)
        dmap = self.df.hvplot.scatter('x', 'y', c='y', **{operation: True})
        agg = dmap.callback.inputs[0].callback.operation.p.aggregator
        self.assertIsInstance(agg, mean)
        self.assertEqual(agg.column, '_color')
        assert original_data.equals(self.df)

    def test_rasterize_color_dim_with_new_column_gets_default_cmap(self):
        plot = self.df.hvplot.scatter('x', 'y', c='y', dynamic=False, rasterize=True)
        opts = Store.lookup_options('bokeh', plot, 'style').kwargs
        self.assertEqual(opts.get('cmap'), 'kbc_r')

    def test_rasterize_default_cmap(self):
        plot = self.df.hvplot.scatter('x', 'y', dynamic=False, rasterize=True)
        opts = Store.lookup_options('bokeh', plot, 'style').kwargs
        self.assertEqual(opts.get('cmap'), 'kbc_r')

    def test_rasterize_set_clim(self):
        plot = self.df.hvplot.scatter('x', 'y', dynamic=False, rasterize=True, clim=(1, 4))
        opts = Store.lookup_options('bokeh', plot, 'plot').kwargs
        self.assertEqual(opts.get('clim'), (1, 4))

    @parameterized.expand([('aspect',), ('data_aspect',)])
    def test_aspect_with_datashade(self, opt):
        plot = self.df.hvplot(x='x', y='y', datashade=True, **{opt: 2})
        opts = Store.lookup_options('bokeh', plot[()], 'plot').kwargs
        self.assertEqual(opts[opt], 2)
        self.assertEqual(opts.get('height'), None)
        self.assertEqual(opts.get('frame_height'), None)

    @parameterized.expand([('aspect',), ('data_aspect',)])
    def test_aspect_with_datashade_and_dynamic_is_false(self, opt):
        plot = self.df.hvplot(x='x', y='y', datashade=True, dynamic=False, **{opt: 2})
        opts = Store.lookup_options('bokeh', plot[()], 'plot').kwargs
        self.assertEqual(opts[opt], 2)
        self.assertEqual(opts.get('height'), None)
        self.assertEqual(opts.get('frame_height'), None)

    @parameterized.expand([('aspect',), ('data_aspect',)])
    def test_aspect_and_frame_height_with_datashade(self, opt):
        plot = self.df.hvplot(x='x', y='y', frame_height=150, datashade=True, **{opt: 2})
        opts = Store.lookup_options('bokeh', plot[()], 'plot').kwargs
        self.assertEqual(opts[opt], 2)
        self.assertEqual(opts.get('frame_height'), 150)
        self.assertEqual(opts.get('height'), None)
        self.assertEqual(opts.get('frame_width'), None)

    @parameterized.expand([('aspect',), ('data_aspect',)])
    def test_aspect_and_frame_height_with_datashade_and_dynamic_is_false(self, opt):
        plot = self.df.hvplot(x='x', y='y', frame_height=150, datashade=True, dynamic=False, **{opt: 2})
        opts = Store.lookup_options('bokeh', plot[()], 'plot').kwargs
        self.assertEqual(opts[opt], 2)
        self.assertEqual(opts.get('frame_height'), 150)
        self.assertEqual(opts.get('height'), None)
        self.assertEqual(opts.get('frame_width'), None)

    def test_cmap_can_be_color_key(self):
        color_key = {'A': '#ff0000', 'B': '#00ff00', 'C': '#0000ff'}
        self.df.hvplot.points(x='x', y='y', by='category', cmap=color_key, datashade=True)
        with self.assertRaises(TypeError):
            self.df.hvplot.points(x='x', y='y', by='category', datashade=True,
                                  cmap='kbc_r', color_key=color_key)

    def test_when_datashade_is_true_set_hover_to_false_by_default(self):
        plot = self.df.hvplot(x='x', y='y', datashade=True)
        opts = Store.lookup_options('bokeh', plot[()], 'plot').kwargs
        assert 'hover' not in opts.get('tools')

    def test_when_datashade_is_true_hover_can_still_be_true(self):
        plot = self.df.hvplot(x='x', y='y', datashade=True, hover=True)
        opts = Store.lookup_options('bokeh', plot[()], 'plot').kwargs
        assert 'hover' in opts.get('tools')

    def test_xlim_affects_x_range(self):
        data = pd.DataFrame(np.random.randn(100).cumsum())
        img = data.hvplot(xlim=(0, 20000), datashade=True, dynamic=False)
        assert img.range(0) == (0, 20000)


class TestChart2D(ComparisonTestCase):

    def setUp(self):
        try:
            import xarray as xr
            import datashader as ds # noqa
        except:
            raise SkipTest('xarray or datashader not available')
        if sys.maxsize < 2**32:
            raise SkipTest('Datashader does not support 32-bit systems')
        import hvplot.xarray  # noqa
        data = np.arange(0, 60).reshape(6, 10)
        x = np.arange(10)
        y = np.arange(6)
        self.da = xr.DataArray(data,
                               coords={'y': y, 'x': x},
                               dims=('y', 'x'))

    @parameterized.expand([('image', Image), ('quadmesh', QuadMesh)])
    def test_plot_resolution(self, kind, element):
        plot = self.da.hvplot(kind=kind)
        assert all(plot.data.x.diff('x').round(0) == 1)
        assert all(plot.data.y.diff('y').round(0) == 1)

    @parameterized.expand([('image', Image), ('quadmesh', QuadMesh)])
    def test_plot_resolution_with_rasterize(self, kind, element):
        plot = self.da.hvplot(kind=kind, dynamic=False, rasterize=True,
                              x_sampling=5, y_sampling=2)
        assert all(plot.data.x.diff('x').round(0) == 5)
        assert all(plot.data.y.diff('y').round(0) == 2)



"""
interactive API
"""

import abc
import operator
import sys

from types import FunctionType, MethodType

import holoviews as hv
import pandas as pd
import panel as pn
import param

from panel.layout import Column, Row, VSpacer, HSpacer
from panel.util import get_method_owner, full_groupby
from panel.widgets.base import Widget

from .util import is_tabular, is_xarray, is_xarray_dataarray


def _find_widgets(op):
    widgets = []
    op_args = list(op['args'])+list(op['kwargs'].values())
    op_args = hv.core.util.flatten(op_args)
    for op_arg in op_args:
        if 'panel' in sys.modules:
            if isinstance(op_arg, Widget) and op_arg not in widgets:
                widgets.append(op_arg)
        if isinstance(op_arg, hv.dim):
            for nested_op in op_arg.ops:
                for widget in _find_widgets(nested_op):
                    if widget not in widgets:
                        widgets.append(widget)
        if 'ipywidgets' in sys.modules:
            from ipywidgets import Widget as IPyWidget
            if isinstance(op_arg, IPyWidget) and op_arg not in widgets:
                widgets.append(op_arg)
        if (isinstance(op_arg, param.Parameter) and
            isinstance(op_arg.owner, pn.widgets.Widget) and
            op_arg.owner not in widgets):
            widgets.append(op_arg.owner)
    return widgets


class Interactive():
    """
    Interactive is a wrapper around a Python object that lets users create
    interactive pipelines by calling existing APIs on an object with
    dynamic parameters or widgets.
    """

    __metaclass__ = abc.ABCMeta

    _fig = None

    def __new__(cls, obj, **kwargs):
        if 'fn' in kwargs:
            fn = kwargs.pop('fn')
        elif isinstance(obj, (FunctionType, MethodType)):
            fn = pn.panel(obj)
            obj = fn.eval(obj)
        else:
            fn = None
        clss = cls
        for subcls in cls.__subclasses__():
            if subcls.applies(obj):
                clss = subcls
        inst = super(Interactive, cls).__new__(clss)
        inst._obj = obj
        inst._fn = fn
        return inst

    @classmethod
    def applies(cls, obj):
        return True

    def __init__(self, obj, transform=None, fn=None, plot=False, depth=0,
                 loc='top_left', center=False, dmap=False, inherit_kwargs={},
                 max_rows=100, method=None, **kwargs):
        self._init = False
        if self._fn is not None:
            for _, params in full_groupby(self._fn_params, lambda x: id(x.owner)):
                params[0].owner.param.watch(self._update_obj, [p.name for p in params])
        self._method = method
        if transform is None:
            dim = '*'
            transform = hv.util.transform.dim
            if is_xarray(obj):
                transform = hv.util.transform.xr_dim
                if is_xarray_dataarray(obj):
                    dim = obj.name
                if dim is None:
                    raise ValueError(
                        "Cannot use interactive API on DataArray without name."
                        "Assign a name to the DataArray and try again."
                    )
            elif is_tabular(obj):
                transform = hv.util.transform.df_dim
            self._transform = transform(dim)
        else:
            self._transform = transform
        self._plot = plot
        self._depth = depth
        self._loc = loc
        self._center = center
        self._dmap = dmap
        self._inherit_kwargs = inherit_kwargs
        self._max_rows = max_rows
        self._kwargs = kwargs
        ds = hv.Dataset(self._obj)
        self._current = self._transform.apply(ds, keep_index=True, compute=False)
        self._init = True

    def _update_obj(self, *args):
        self._obj = self._fn.eval(self._fn.object)

    @property
    def _fn_params(self):
        if self._fn is None:
            deps = []
        elif isinstance(self._fn, pn.param.ParamFunction):
            dinfo = getattr(self._fn.object, '_dinfo', {})
            deps = list(dinfo.get('dependencies', [])) + list(dinfo.get('kw', {}).values())
        else:
            parameterized = get_method_owner(self._fn.object)
            deps = parameterized.param.method_dependencies(self._fn.object.__name__)
        return deps
        
    @property
    def _params(self):
        ps = self._fn_params
        for k, p in self._transform.params.items():
            if k == 'ax' or p in ps:
                continue
            ps.append(p)
        return ps

    @property
    def _callback(self):
        @pn.depends(*self._params)
        def evaluate(*args, **kwargs):
            obj = self._obj
            ds = hv.Dataset(obj)
            transform = self._transform
            if ds.interface.datatype == 'xarray' and is_xarray_dataarray(obj):
                transform = transform.clone(obj.name)
            obj = transform.apply(ds, keep_index=True, compute=False)
            if self._method:
                obj = getattr(obj, self._method, obj)
            if self._plot:
                return Interactive._fig
            elif isinstance(obj, pd.DataFrame):
                return pn.pane.DataFrame(obj, max_rows=self._max_rows, **self._kwargs)
            else:
                return obj
        return evaluate

    def _clone(self, transform=None, plot=None, loc=None, center=None,
               dmap=None, copy=False, **kwargs):
        plot = self._plot or plot
        transform = transform or self._transform
        loc = self._loc if loc is None else loc
        center = self._center if center is None else center
        dmap = self._dmap if dmap is None else dmap
        depth = self._depth+1
        if copy:
            kwargs = dict(self._kwargs, inherit_kwargs=self._inherit_kwargs, method=self._method, **kwargs)
        else:
            kwargs = dict(self._inherit_kwargs, **dict(self._kwargs, **kwargs))
        return type(self)(self._obj, fn=self._fn, transform=transform, plot=plot, depth=depth,
                         loc=loc, center=center, dmap=dmap, **kwargs)

    def _repr_mimebundle_(self, include=[], exclude=[]):
        return self.layout()._repr_mimebundle_()

    def __dir__(self):
        current = self._current
        if self._method:
            current = getattr(current, self._method)
        extras = {attr for attr in dir(current) if not attr.startswith('_')}
        if is_tabular(current) and hasattr(current, 'columns'):
            extras |= set(current.columns)
        try:
            return sorted(set(super().__dir__()) | extras)
        except Exception:
            return sorted(set(dir(type(self))) | set(self.__dict__) | extras)

    def _resolve_accessor(self):
        if not self._method:
            return self._clone(copy=True)
        transform = type(self._transform)(self._transform, self._method, accessor=True)
        transform._ns = self._current
        inherit_kwargs = {}
        if self._method == 'plot':
            inherit_kwargs['ax'] = self._get_ax_fn()
        try:
            new = self._clone(transform, inherit_kwargs=inherit_kwargs)
        finally:
            self._method = None
        return new

    def __getattribute__(self, name):
        self_dict = super().__getattribute__('__dict__')
        if not self_dict.get('_init'):
            return super().__getattribute__(name)

        current = self_dict['_current']
        method = self_dict['_method']
        if method:
            current = getattr(current, method)
        extras = [d for d in dir(current) if not d.startswith('_')]
        if name in extras and name not in super().__dir__():
            new = self._resolve_accessor()
            new._method = name
            try:
                new.__doc__ = getattr(current, name).__doc__
            except Exception:
                pass
            return new
        return super().__getattribute__(name)

    @staticmethod
    def _get_ax_fn():
        @pn.depends()
        def get_ax():
            from matplotlib.backends.backend_agg import FigureCanvas
            from matplotlib.pyplot import Figure
            Interactive._fig = fig = Figure()
            FigureCanvas(fig)
            return fig.subplots()
        return get_ax

    def __call__(self, *args, **kwargs):
        if self._method is None:
            if self._depth == 0:
                return self._clone(*args, **kwargs)
            raise AttributeError
        elif self._method == 'plot':
            kwargs['ax'] = self._get_ax_fn()
        new = self._clone(copy=True)
        try:
            method = type(new._transform)(new._transform, new._method, accessor=True)
            kwargs = dict(new._inherit_kwargs, **kwargs)
            clone = new._clone(method(*args, **kwargs), plot=new._method == 'plot')
        finally:
            new._method = None
        return clone

    #----------------------------------------------------------------
    # Interactive pipeline APIs
    #----------------------------------------------------------------

    def __array_ufunc__(self, *args, **kwargs):
        new = self._resolve_accessor()
        transform = new._transform
        transform = args[0](transform, *args[3:], **kwargs)
        return new._clone(transform)

    def _apply_operator(self, operator, *args, **kwargs):
        new = self._resolve_accessor()
        transform = new._transform
        transform = type(transform)(transform, operator, *args)
        return new._clone(transform)

    # Builtin functions
    def __abs__(self):
        return self._apply_operator(abs)

    def __round__(self, ndigits=None):
        args = () if ndigits is None else (ndigits,)
        return self._apply_operator(round, *args)

    # Unary operators
    def __neg__(self):
        return self._apply_operator(operator.neg)
    def __not__(self):
        return self._apply_operator(operator.not_)
    def __invert__(self):
        return self._apply_operator(operator.inv)
    def __pos__(self):
        return self._apply_operator(operator.pos)

    # Binary operators
    def __add__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.add, other)
    def __and__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.and_, other)
    def __div__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.div, other)
    def __eq__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.eq, other)
    def __floordiv__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.floordiv, other)
    def __ge__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.ge, other)
    def __gt__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.gt, other)
    def __le__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.le, other)
    def __lt__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.lt, other)
    def __lshift__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.lshift, other)
    def __mod__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.mod, other)
    def __mul__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.mul, other)
    def __ne__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.ne, other)
    def __or__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.or_, other)
    def __rshift__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.rshift, other)
    def __pow__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.pow, other)
    def __sub__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.sub, other)
    def __truediv__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.truediv, other)

    # Reverse binary operators
    def __radd__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.div, other, reverse=True)
    def __rand__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.and_, other, reverse=True)
    def __rdiv__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.div, other, reverse=True)
    def __rfloordiv__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.floordiv, other, reverse=True)
    def __rlshift__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.rlshift, other)
    def __rmod__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.mod, other, reverse=True)
    def __rmul__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.mul, other, reverse=True)
    def __ror__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.or_, other, reverse=True)
    def __rpow__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.pow, other, reverse=True)
    def __rrshift__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.rrshift, other)
    def __rsub__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.sub, other, reverse=True)
    def __rtruediv__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.truediv, other, reverse=True)

    def __getitem__(self, other):
        other = other._transform if isinstance(other, Interactive) else other
        return self._apply_operator(operator.getitem, other)

    def _plot(self, *args, **kwargs):
        @pn.depends()
        def get_ax():
            from matplotlib.backends.backend_agg import FigureCanvas
            from matplotlib.pyplot import Figure
            Interactive._fig = fig = Figure()
            FigureCanvas(fig)
            return fig.subplots()
        kwargs['ax'] = get_ax
        new = self._resolve_accessor()
        transform = new._transform
        transform = type(transform)(transform, 'plot', accessor=True)
        return new._clone(transform(*args, **kwargs), plot=True)

    def hvplot(self, *args, **kwargs):
        new = self._resolve_accessor()
        transform = new._transform
        transform = type(transform)(transform, 'hvplot', accessor=True)
        dmap = 'kind' not in kwargs
        return new._clone(transform(*args, **kwargs), dmap=dmap)

    #----------------------------------------------------------------
    # Public API
    #----------------------------------------------------------------

    def dmap(self):
        """
        Wraps the output in a DynamicMap. Only valid if the output
        is a HoloViews object.
        """
        return hv.DynamicMap(self._callback)

    def eval(self):
        """
        Returns the currrent state of the interactive expression. The
        returned object is no longer interactive.
        """
        obj = self._current
        if self._method:
            return getattr(obj, self._method, obj)
        return obj

    def layout(self, **kwargs):
        """
        Returns a layout of the widgets and output arranged according
        to the center and widget location specified in the
        interactive call.
        """
        widget_box = self.widgets()
        panel = self.output()
        loc = self._loc
        if loc in ('left', 'right'):
            widgets = Column(VSpacer(), widget_box, VSpacer())
        elif loc in ('top', 'bottom'):
            widgets = Row(HSpacer(), widget_box, HSpacer())
        elif loc in ('top_left', 'bottom_left'):
            widgets = Row(widget_box, HSpacer())
        elif loc in ('top_right', 'bottom_right'):
            widgets = Row(HSpacer(), widget_box)
        elif loc in ('left_top', 'right_top'):
            widgets = Column(widget_box, VSpacer())
        elif loc in ('left_bottom', 'right_bottom'):
            widgets = Column(VSpacer(), widget_box)
        center = self._center
        if not widgets:
            if center:
                components = [HSpacer(), panel, HSpacer()]
            else:
                components = [panel]
        elif center:
            if loc.startswith('left'):
                components = [widgets, HSpacer(), panel, HSpacer()]
            elif loc.startswith('right'):
                components = [HSpacer(), panel, HSpacer(), widgets]
            elif loc.startswith('top'):
                components = [HSpacer(), Column(widgets, Row(HSpacer(), panel, HSpacer())), HSpacer()]
            elif loc.startswith('bottom'):
                components = [HSpacer(), Column(Row(HSpacer(), panel, HSpacer()), widgets), HSpacer()]
        else:
            if loc.startswith('left'):
                components = [widgets, panel]
            elif loc.startswith('right'):
                components = [panel, widgets]
            elif loc.startswith('top'):
                components = [Column(widgets, panel)]
            elif loc.startswith('bottom'):
                components = [Column(panel, widgets)]
        return Row(*components, **kwargs)

    def holoviews(self):
        """
        Returns a HoloViews object to render the output of this
        pipeline. Only works if the output of this pipeline is a
        HoloViews object, e.g. from an .hvplot call.
        """
        return hv.DynamicMap(self._callback)

    def output(self):
        """
        Returns the output of the interactive pipeline, which is
        either a HoloViews DynamicMap or a Panel object.

        Returns
        -------
        DynamicMap or Panel object wrapping the interactive output.
        """
        return self.holoviews() if self._dmap else self.panel(**self._kwargs)

    def panel(self, **kwargs):
        """
        Wraps the output in a Panel component.
        """
        return pn.panel(self._callback, **kwargs)

    def widgets(self):
        """
        Returns a Column of widgets which control the interactive output.

        Returns
        -------
        A Column of widgets
        """
        widgets = []
        for p in self._fn_params:
            if (isinstance(p.owner, pn.widgets.Widget) and
                p.owner not in widgets):
                widgets.append(p.owner)
        for op in self._transform.ops:
            for w in _find_widgets(op):
                if w not in widgets:
                    widgets.append(w)
        return pn.Column(*widgets)
